/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2017                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Implementation of the class MutableGlobalDefs.
 *
 *  This file provides the implementation of the class MutableGlobalDefs.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "MutableGlobalDefs.h"

#include <cfloat>

#include <UTILS_Error.h>

#include <pearl/Error.h>

#include "CallTree.h"
#include "DefinitionContainer.h"
#include "DefinitionTree.h"

using namespace std;
using namespace pearl;
using namespace pearl::detail;


// --- Helper functions -----------------------------------------------------

namespace
{
/// @brief   Ensures unique IDs for global definitions.
/// @relates pearl::detail::MutableGlobalDefs
///
/// Checks whether the global definition objects stored in @p container have
/// unique global identifiers and throws an exception if not.  This function
/// implicitly also optimizes lookups by identifier.
///
/// @param container
///     Global definition container
/// @param defName
///     Definition name used in the exception message
///
/// @tparam DefContainerT
///     Global definition container type providing the required typedef's and
///     member functions (DefinitionContainer or DefinitionTree)
///
/// @throws RuntimeError
///     on identifying a duplicate identifier
///
template< typename DefContainerT >
void
ensureUniqueIds(DefContainerT& container,
                const string&  defName)
{
    // Special case: empty container
    if (container.isEmpty())
    {
        return;
    }

    // Step 1: ensure that definitions are sorted by ID
    container.optimizeLookups();

    // Step 2: ensure that consecutive definitions have different IDs
    typename DefContainerT::ElementType::IdType prevId =
        container.getItemByIndex(0)->getId();
    const size_t numItems = container.numItems();
    for (size_t i = 1; i < numItems; ++i)
    {
        typename DefContainerT::ElementType::IdType currId =
            container.getItemByIndex(i)->getId();
        if (currId == prevId)
        {
            throw RuntimeError("pearl::detail::MutableGlobalDefs::"
                               "validate() -- "
                               "duplicate " + defName + " ID: "
                               + toStdString(currId));
        }
        prevId = currId;
    }
}
}    // unnamed namespace


// --- Constructors & destructor --------------------------------------------

MutableGlobalDefs::MutableGlobalDefs()
    : mGlobalOffset(DBL_MAX),
      mLastDefType(UNDEFINED)
{
}


// --- Insertion functions for global definitions ---------------------------

void
MutableGlobalDefs::addCallingContext(CallingContext* const item)
{
    trackType(CALLINGCONTEXT);

    mContextTree->addItem(item);
}


/// @todo Check exceptions thrown on error
void
MutableGlobalDefs::addCallpath(Callpath* const item)
{
    trackType(CALLPATH);

    mCallTree->addCallpath(item);
}


void
MutableGlobalDefs::addCallsite(Callsite* const item)
{
    trackType(CALLSITE);

    mCallsites->addItem(item);
}


void
MutableGlobalDefs::addCartesian(Cartesian* const item)
{
    trackType(CARTESIAN);

    mCartesians->addItem(item);
}


void
MutableGlobalDefs::addCommunicator(Communicator* const item)
{
    trackType(COMMUNICATOR);

    mCommunicators->addItem(item);
}


void
MutableGlobalDefs::addLocation(Location* const item)
{
    trackType(LOCATION);

    mLocations->addItem(item);
}


void
MutableGlobalDefs::addLocationGroup(LocationGroup* const item)
{
    trackType(LOCATIONGROUP);

    mLocationGroups->addItem(item);
}


void
MutableGlobalDefs::addMetric(Metric* const item)
{
    trackType(METRIC);

    mMetrics->addItem(item);
}


void
MutableGlobalDefs::addProcessGroup(ProcessGroup* const item)
{
    trackType(PROCESSGROUP);

    mProcessGroups->addItem(item);
}


void
MutableGlobalDefs::addRegion(Region* const item)
{
    trackType(REGION);

    mRegions->addItem(item);
}


void
MutableGlobalDefs::addRmaWindow(RmaWindow* const item)
{
    trackType(RMAWINDOW);

    mRmaWindows->addItem(item);
}


void
MutableGlobalDefs::addSourceLocation(SourceLocation* const item)
{
    trackType(SOURCELOCATION);

    mSourceLocations->addItem(item);
}


void
MutableGlobalDefs::addString(String* const item)
{
    trackType(STRING);

    mStrings->addItem(item);
}


void
MutableGlobalDefs::addSystemNode(SystemNode* const item)
{
    trackType(SYSTEMNODE);

    mSystemTree->addItem(item);
}


// --- Internal methods -----------------------------------------------------

void
MutableGlobalDefs::validate()
{
    ensureUniqueIds(*mContextTree, "CallingContext");
    ensureUniqueIds(*mCallsites, "Callsite");
    ensureUniqueIds(*mCartesians, "Cartesian");
    ensureUniqueIds(*mCommunicators, "Communicator");
    ensureUniqueIds(*mLocations, "Location");
    ensureUniqueIds(*mLocationGroups, "LocationGroup");
    ensureUniqueIds(*mMetrics, "Metric");
    ensureUniqueIds(*mProcessGroups, "ProcessGroup");
    ensureUniqueIds(*mRegions, "Region");
    ensureUniqueIds(*mRmaWindows, "RmaWindow");
    ensureUniqueIds(*mSourceLocations, "SourceLocation");
    ensureUniqueIds(*mStrings, "String");
    ensureUniqueIds(*mSystemTree, "SystemNode");
}


/// @todo Check whether UNKNOWN callpath is still necessary
/// @todo Refactor, then remove optimizeLookups() calls already triggered by
///       validate()
void
MutableGlobalDefs::setup()
{
    mContextTree->optimizeLookups();
    mCallsites->optimizeLookups();
    mCartesians->optimizeLookups();
    mCommunicators->optimizeLookups();
    mLocations->optimizeLookups();
    mLocationGroups->optimizeLookups();
    mMetrics->optimizeLookups();
    mProcessGroups->optimizeLookups();
    mRegions->optimizeLookups();
    mRmaWindows->optimizeLookups();
    mSourceLocations->optimizeLookups();
    mStrings->optimizeLookups();
    mSystemTree->optimizeLookups();

    // Determine next String identifier
    String::IdType nextStringId = 0;
    if (!mStrings->isEmpty())
    {
        const size_t lastIndex = mStrings->numItems() - 1;

        nextStringId = mStrings->getItemByIndex(lastIndex)->getId() + 1;
    }

    // Determine next Region identifier
    Region::IdType nextRegionId = 0;
    if (!mRegions->isEmpty())
    {
        const size_t lastIndex = mRegions->numItems() - 1;

        nextRegionId = mRegions->getItemByIndex(lastIndex)->getId() + 1;
    }

    // Create special "MEASUREMENT OFF" region if necessary
    if (!mPausingRegion)
    {
        // Create "MEASUREMENT OFF" string
        String* pauseStr = new String(nextStringId++, "MEASUREMENT OFF");
        addString(pauseStr);

        // Create "MEASUREMENT OFF" region
        mPausingRegion = new Region(nextRegionId++,
                                    *pauseStr,
                                    *pauseStr,
                                    String::UNDEFINED,
                                    Region::ROLE_ARTIFICIAL,
                                    Paradigm::USER,
                                    String::UNDEFINED,
                                    PEARL_NO_NUM,
                                    PEARL_NO_NUM);
        addRegion(mPausingRegion);
    }

    // Create special "TRACE BUFFER FLUSH" region if necessary
    if (!mFlushingRegion)
    {
        // Create "TRACE BUFFER FLUSH" string
        String* flushStr = new String(nextStringId++, "TRACE BUFFER FLUSH");
        addString(flushStr);

        // Create "TRACE BUFFER FLUSH" region
        mFlushingRegion = new Region(nextRegionId++,
                                     *flushStr,
                                     *flushStr,
                                     String::UNDEFINED,
                                     Region::ROLE_ARTIFICIAL,
                                     Paradigm::MEASUREMENT_SYSTEM,
                                     String::UNDEFINED,
                                     PEARL_NO_NUM,
                                     PEARL_NO_NUM);
        addRegion(mFlushingRegion);
    }

    // Create special "TASKS" region if necessary
    if (!mTaskRootRegion)
    {
        // Create "TASKS" string
        String* taskRootStr = new String(nextStringId++, "TASKS");
        addString(taskRootStr);

        // Create "TASKS" region
        mTaskRootRegion = new Region(nextRegionId++,
                                     *taskRootStr,
                                     *taskRootStr,
                                     String::UNDEFINED,
                                     Region::ROLE_ARTIFICIAL,
                                     Paradigm::MEASUREMENT_SYSTEM,
                                     String::UNDEFINED,
                                     PEARL_NO_NUM,
                                     PEARL_NO_NUM);
        addRegion(mTaskRootRegion);
    }

    // Create special "THREADS" region if necessary
    if (!mThreadRootRegion)
    {
        // Create "THREADS" string
        String* threadRootStr = new String(nextStringId++, "THREADS");
        addString(threadRootStr);

        // Create "THREADS" region
        mThreadRootRegion = new Region(nextRegionId++,
                                       *threadRootStr,
                                       *threadRootStr,
                                       String::UNDEFINED,
                                       Region::ROLE_ARTIFICIAL,
                                       Paradigm::MEASUREMENT_SYSTEM,
                                       *threadRootStr,
                                       PEARL_NO_NUM,
                                       PEARL_NO_NUM);
        addRegion(mThreadRootRegion);
    }

    /*
    // Create "UNKNOWN" call path
    if (mCalltree->numCallpaths() == 0)
    mCalltree->getCallpath(Region::UNDEFINED, Callsite::UNDEFINED, NULL);
    */

    // Reset call tree modification flag
    mCallTree->setModified(false);
}


// --- Set internal regions -------------------------------------------------

void
MutableGlobalDefs::setPausingRegion(Region* const region)
{
    if (mPausingRegion)
    {
        throw FatalError("Duplicate definition of PAUSING region!");
    }

    mPausingRegion = region;
}


void
MutableGlobalDefs::setFlushingRegion(Region* const region)
{
    if (mFlushingRegion)
    {
        throw FatalError("Duplicate definition of FLUSHING region!");
    }

    mFlushingRegion = region;
}


void
MutableGlobalDefs::setTaskRootRegion(Region* const region)
{
    if (mTaskRootRegion)
    {
        throw FatalError("Duplicate definition of Task Root region!");
    }

    mTaskRootRegion = region;
}


void
MutableGlobalDefs::setThreadRootRegion(Region* const region)
{
    if (mThreadRootRegion)
    {
        throw FatalError("Duplicate definition of THREADS Root region!");
    }

    mThreadRootRegion = region;
}


// --- Call tree handling ---------------------------------------------------

CallTree*
MutableGlobalDefs::getCallTree() const
{
    return mCallTree.get();
}


void
MutableGlobalDefs::setCallTree(CallTree* const callTree)
{
    if (mCallTree.get() == callTree)
    {
        return;
    }

    mCallTree.reset(callTree);
}


// --- Global time offset handling ------------------------------------------

timestamp_t
MutableGlobalDefs::getGlobalOffset() const
{
    return mGlobalOffset;
}


void
MutableGlobalDefs::setGlobalOffset(const timestamp_t offset)
{
    mGlobalOffset = offset;
}


// --- Private methods ------------------------------------------------------

/// @brief Track definition type and optimize lookups for previous type.
///
/// Tracks the provided type @p currDefType of the current definition.  If
/// it differs from the type of the definition seen previously, optimizes
/// item lookups by global identifier for objects of the previous type.
///
/// @param currDefType
///     Type of the current definition
///
/// @exception_nothrow
///
void
MutableGlobalDefs::trackType(const DefinitionType currDefType)
{
    if (currDefType != mLastDefType)
    {
        // Optimize lookups for definitions of the previous type
        switch (mLastDefType)
        {
            case UNDEFINED:
                // This is the initial value, i.e., we get here only for the
                // first definition object added.  In this case, there is
                // nothing to be optimized yet.
                break;

            case CALLINGCONTEXT:
                mContextTree->optimizeLookups();
                break;

            case CALLPATH:
                // Call trees don't use the DefinitionTree class yet...
                break;

            case CALLSITE:
                mCallsites->optimizeLookups();
                break;

            case CARTESIAN:
                mCartesians->optimizeLookups();
                break;

            case COMMUNICATOR:
                mCommunicators->optimizeLookups();
                break;

            case LOCATION:
                mLocations->optimizeLookups();
                break;

            case LOCATIONGROUP:
                mLocationGroups->optimizeLookups();
                break;

            case METRIC:
                mMetrics->optimizeLookups();
                break;

            case PROCESSGROUP:
                mProcessGroups->optimizeLookups();
                break;

            case REGION:
                mRegions->optimizeLookups();
                break;

            case RMAWINDOW:
                mRmaWindows->optimizeLookups();
                break;

            case SOURCELOCATION:
                mSourceLocations->optimizeLookups();
                break;

            case STRING:
                mStrings->optimizeLookups();
                break;

            case SYSTEMNODE:
                mSystemTree->optimizeLookups();
                break;

            default:
                UTILS_BUG("Unknown definition type");
        }

        // Track definition type
        mLastDefType = currDefType;
    }
}
